/*
 ***********************************************************************************************
 *
 *  OSA cooperative RTOS for microcontrollers PIC (Microchip) and 8-bit AVR (Atmel)
 *
 *  OSA is distributed under BSD license (see license.txt)
 *
 *  URL:        http://wiki.pic24.ru/doku.php/en/osa/ref/intro
 *              http://picosa.narod.ru
 *
 *----------------------------------------------------------------------------------------------
 *
 *  File:       osa_bsem.h
 *
 *  Programmer: Timofeev Victor
 *              osa@pic24.ru, testerplus@mail.ru
 *
 *  Definition: Services for work with binary semaphores
 *
 *  History:    21.01.2009
 * 
 *              10.12.2009      1. Services for interrupt were corrected.
 *                                  Modified:
 *                                      OS_Bsem_Wait
 *                                      OS_Bsem_Wait_TO
 *                                  Added:
 *                                      OS_Bsem_Check_I
 *                                      OS_Bsem_Reset_I
 *                                      OS_Bsem_Set_I
 *                                      OS_Bsem_Switch_I
 *                                      OS_Bsem_Signal_I
 * 
 *
 ************************************************************************************************
 */

/************************************************************************************************
 *                                                                                              *
 *                          B I N A R Y    S E M A P H O R E S                                  *
 *                                                                                              *
 ************************************************************************************************/


#if defined(OS_ENABLE_INT_BSEM)

    #define __OS_BSEM_DI()          OS_DI_INT()
    #define __OS_BSEM_RI()          OS_RI_INT()

    #if defined(OS_PROTECT_MEMORY_ACCESS)           /*100312.1*/
        #define __OS_BSEM_DI_A()    OS_DI_INT()
        #define __OS_BSEM_RI_A()    OS_RI_INT()
    #else
        #define __OS_BSEM_DI_A()
        #define __OS_BSEM_RI_A()
    #endif

#else

    #define __OS_BSEM_DI()
    #define __OS_BSEM_RI()
    #define __OS_BSEM_DI_A()                        /*100312.1*/
    #define __OS_BSEM_RI_A()

#endif


#define OS_BSEM_ATOMIC_WRITE_A(expr)    \
    OSM_BEGIN {                         \
        __OS_BSEM_DI_A();               \
        expr;                           \
        __OS_BSEM_RI_A();               \
    } OSM_END


#define OS_BSEM_ATOMIC_READ_A(expr)                                     \
    (__OS_BSEM_DI_A(), _OS_temp = (expr), __OS_BSEM_RI_A(), _OS_temp)   \

//------------------------------------------------------------------------------


#define _OS_BSEM_POS(binsem)        ((binsem) >> _OST_INT_SHIFT)
#define _OS_BSEM_MASK(binsem)       (1 << ((binsem) & _OST_INT_MASK))



#define __OS_Bsem_Set(binsem)       OS_BSem[_OS_BSEM_POS(binsem)] |=  _OS_BSEM_MASK(binsem) /*100312.1*/
#define __OS_Bsem_Reset(binsem)     OS_BSem[_OS_BSEM_POS(binsem)] &= ~_OS_BSEM_MASK(binsem) /*100312.1*/
#define __OS_Bsem_Switch(binsem)    OS_BSem[_OS_BSEM_POS(binsem)] ^=  _OS_BSEM_MASK(binsem) /*100312.1*/
#define __OS_Bsem_Check(binsem)    (OS_BSem[_OS_BSEM_POS(binsem)] &   _OS_BSEM_MASK(binsem))/*100312.1*/


// Set/reset/switch  binary semaphore

#define OS_Bsem_Set(binsem)           OS_BSEM_ATOMIC_WRITE_A(__OS_Bsem_Set(binsem))       /*100312.1*/
#define OS_Bsem_Reset(binsem)         OS_BSEM_ATOMIC_WRITE_A(__OS_Bsem_Reset(binsem))     /*100312.1*/
#define OS_Bsem_Signal(binsem)        OS_BSEM_ATOMIC_WRITE_A(__OS_Bsem_Set(binsem))       /*100312.1*/
#define OS_Bsem_Switch(binsem)        OS_BSEM_ATOMIC_WRITE_A(__OS_Bsem_Switch(binsem))    /*100312.1*/

// Check for binary semaphore is set

#define OS_Bsem_Check(binsem)   __OS_Bsem_Check(binsem)


// Inverce semaphore's value



//------------------------------------------------------------------------------
#if defined(OS_ENABLE_INT_BSEM)
//------------------------------------------------------------------------------

    // Wait binary semaphore

    #define OS_Bsem_Wait(binsem)                                            \
        OSM_BEGIN {                                                         \
            for (;;) {                                                      \
                _OS_WAIT_EVENT(OS_Bsem_Check(binsem));                      \
                __OS_BSEM_DI();                                             \
                if (OS_Bsem_Check(binsem)) break;                           \
                __OS_BSEM_DI();                                             \
            }                                                               \
            __OS_Bsem_Reset(binsem);                                        \
            __OS_BSEM_RI();                                                 \
        } OSM_END



    // Wait binary semaphore. Exit if timeout expired

    #define OS_Bsem_Wait_TO(binsem, timeout)                                \
        OSM_BEGIN {                                                         \
            for (;;) {                                                      \
                _OS_WAIT_EVENT_TIMEOUT(OS_Bsem_Check(binsem), timeout);     \
                __OS_BSEM_DI();                                             \
                if (OS_Bsem_Check(binsem) || OS_IsTimeout()) break;         \
                __OS_BSEM_RI();                                             \
            }                                                               \
            __OS_Bsem_Reset(binsem);                                        \
            __OS_BSEM_RI();                                                 \
        } OSM_END


    #define OS_Bsem_Set_I(binsem)           __OS_Bsem_Set(binsem)       /*100312.1*/
    #define OS_Bsem_Reset_I(binsem)         __OS_Bsem_Reset(binsem)     /*100312.1*/
    #define OS_Bsem_Signal_I(binsem)        __OS_Bsem_Set(binsem)       /*100312.1*/
    #define OS_Bsem_Switch_I(binsem)        __OS_Bsem_Switch(binsem)    /*100312.1*/
    #define OS_Bsem_Check_I(binsem)         __OS_Bsem_Check(binsem)     /*100312.1*/


//------------------------------------------------------------------------------
#else   // OS_ENABLE_INT_BSEM
//------------------------------------------------------------------------------

    // Wait binary semaphore

    #define OS_Bsem_Wait(binsem)                                            \
        OSM_BEGIN {                                                         \
            _OS_WAIT_EVENT(OS_Bsem_Check(binsem));                          \
            __OS_Bsem_Reset(binsem);                                        \
        } OSM_END



    // Wait binary semaphore. Exit if timeout expired

    #define OS_Bsem_Wait_TO(binsem, timeout)                                \
        OSM_BEGIN {                                                         \
            _OS_WAIT_EVENT_TIMEOUT(OS_Bsem_Check(binsem), timeout);         \
            __OS_Bsem_Reset(binsem);                                        \
        } OSM_END

//------------------------------------------------------------------------------
#endif  // OS_ENABLE_INT_BSEM
//------------------------------------------------------------------------------




















