/*
 ************************************************************************************************
 *                                                                                              *
 *  OSA cooperative RTOS for microcontrollers PIC (Microchip) and 8-bit AVR (Atmel)
 *                                                                                              *
 *  URL:            http://wiki.pic24.ru/doku.php/en/osa/ref/intro                              *
 *                  http://picosa.narod.ru                                                      *
 *                                                                                              *
 *----------------------------------------------------------------------------------------------*
 *                                                                                              *
 *  File:           osa_ftimer.c                                                                *    
 *                                                                                              *
 *  Compilers:      HT-PICC STD                                                                 *
 *                  HT-PICC18 STD                                                               *
 *                  Microchip C18                                                               *
 *                  Microchip C30                                                               *
 *                                                                                              *
 *  Programmer:     Timofeev Victor                                                             *
 *                  osa@pic24.ru, testerplus@mail.ru                                            *
 *                                                                                              *
 *  Description:                                                                                *
 *                                                                                              *
 *  History:        14.01.2010 -    File created                                                *
 *                                                                                              *
 ************************************************************************************************
 */


//******************************************************************************
//  
//******************************************************************************


/*
 ********************************************************************************
 *
 *  _OS_Qtimer_List
 *
 *------------------------------------------------------------------------------
 *
 *  description:    (Internal function called by system kernel througth service
 *                  OS)
 *
 *                  This function works with timer's list. It adds, deletes,
 *                  counts timers according to value of ftimer->Flags.
 * 
 *  parameters:     _OST_QTIMER time   - new time for added timer (when
 *                                      _OS_QTimers.Flags.bListFunction =
 *                                      FUNCTION_ADD)
 *                  OST_QTIMER *ftimer - pointer to timer to work on.
 *
 *  on return:      _OST_QTIMER:
 *                  1.  When bListGetTime == "1", this function returns remaining
                        time for current timer (flag bListFunction does not matter)
                    2.  When bListFunction == OS_QTIMER_LIST_FUNCTION_ADD, this
                        function adds new timer into list (before adding it checks
                        for timer is already in list; if so, then timer deleted first
                        and function executes twice (see "goto RESTART"))
                    3.  When bListFunction == OS_QTIMER_LIST_FUNCTION_DELETE, this
                        function deletes current timer
 
 *  Overloaded in:
 * 
 ********************************************************************************
 */

#ifdef OS_ENABLE_QTIMERS

//******************************************************************************
#if defined(__OSAPICC16__)
//******************************************************************************


    _OST_QTIMER _OS_Qtimer_List (_OST_QTIMER time, OST_QTIMER *ftimer)
    {
        OST_QTIMER_FLAGS    flags;
        _OST_QTIMER         temp, remain;
        unsigned char       _temp_i;        // For GIE save
    
    RESTART:
        _fsr = (_FSR_TYPE)ftimer;           //flags = ftimer->Flags;
        _irp = 0;
        *(unsigned char*)&flags = _indf;
    
    
    
        if (_OS_QTimers.Flags.bListGetTime) 
        {    
            //--------------------------------------------
            // Clear ADD function if GET function present
            //--------------------------------------------
            _OS_QTimers.Flags.bListFunction = !OS_QTIMER_LIST_FUNCTION_ADD;
        }
        else 
        {
            //---------------------------------------
            // Skip deleting if timer is not in list
            //---------------------------------------
            if (_OS_QTimers.Flags.bListFunction == OS_QTIMER_LIST_FUNCTION_DELETE && !flags.bActive) goto EXIT;
        }
    
        _OS_QTimers.Flags.bAddAfterDelete = 0;
    
        if (_OS_QTimers.Flags.bListFunction == OS_QTIMER_LIST_FUNCTION_ADD)
        {
            if (flags.bActive)              // If timer is already in list, it should be deleted first
            {
                _OS_QTimers.Flags.bAddAfterDelete = 1;  // Set flag to execute ADD after deleting
                _OS_QTimers.Flags.bListFunction = OS_QTIMER_LIST_FUNCTION_DELETE;
            }
            else
            {
                //---------------------------------------
                // Prepare timer for adding
                //---------------------------------------
                
                _OS_IndfFTimerNextEnable = 0;       //ftimer->Flags.bNextEnable = 0;
                time ^= -1; // time = -time
                time++;
            }
        }
        remain = 0;                                 // Count remaining time
    
        _fsr = (_FSR_TYPE)&_OS_QTimers;             // Start search from header
        // irp is already cleared
    
        _temp_i = _intcon; 
        di();
    
        if (_OS_QTimers.Flags.bNextEnable)
        {    
            do
            {
                _fsr ++;    // Now FSR points to ftcb->next

                if (_OS_QTimers.Flags.bListFunction == OS_QTIMER_LIST_FUNCTION_DELETE && !_OS_QTimers.Flags.bListGetTime) 
                {
                    if (_indf == (unsigned char)ftimer)     //********* Deleting ***************
                    {
                        _fsr--;
                        _OS_IndfFTimerNextEnable = 0;
                        if (flags.bNextEnable) _OS_IndfFTimerNextEnable = 1;

                        _fsr++;
                        *((unsigned char*)&temp + 0) = _fsr;    // Save address of ftcb->Next
                        _fsr = _indf;
                        _OS_IndfFTimerActive = 0;               // ftimer->Falgs.bActive = 0;
                        _fsr++;
                        *((unsigned char*)&remain + 0) = _indf; // Save ftimer->Next

                        if (flags.bNextEnable)                  // Save ftimer->Timer
                        {    
                            _fsr++;

                            *((unsigned char *)&time + 0) = _indf;
                        #if OS_QTIMER_SIZE > 1
                            _fsr++; *((unsigned char*)&time + 1) = _indf;
                        #if OS_QTIMER_SIZE > 2
                            _fsr++; *((unsigned char*)&time + 2) = _indf;
                            _fsr++; *((unsigned char*)&time + 3) = _indf;
                        #endif
                        #endif
                        }

                        _fsr = *((unsigned char*)&temp + 0);    // Restore address of ftcb->Next
                        _indf = *((unsigned char*)&remain + 0); // ftcb->Next = ftimer->Next

                        if (flags.bNextEnable)                  // Add ftimer->Timer to ftcb->Timer
                        {    
                            _fsr = _indf + 2;                       

                            _indf += *((unsigned char*)&time + 0);   
                        #if OS_QTIMER_SIZE == 2
                            _fsr++;
                            if (CARRY) _indf++;
                            _indf += *((unsigned char*)&time + 1);
                        #endif
                        #if OS_QTIMER_SIZE == 4
                            if (CARRY) time -= 0x100;
                            _fsr++;
                            _indf += *((unsigned char *)&time + 1);
                            if (CARRY) time -= 0x10000;
                            _fsr++;
                            _indf += *((unsigned char *)&time + 2);
                            if (CARRY) time -= 0x1000000;
                            _fsr++;
                            _indf += *((unsigned char *)&time + 3);
                        #endif
                        }

                        if (_OS_QTimers.Flags.bAddAfterDelete)
                        {
                            //-----------------------------------------------------------
                            // It was deleting before adding. 
                            // Set ADD flag and restart searching
                            //-----------------------------------------------------------

                            _OS_QTimers.Flags.bListFunction = OS_QTIMER_LIST_FUNCTION_ADD;
                            goto RESTART;
                        }
                        goto EXIT_RI;

                    }   // Deleting: if frs->Next = ftimer
                }
                else   // _OS_QTimers.Flags.bListFunction == OS_QTIMER_LIST_FUNCTION_DELETE && !_OS_QTimers.Flags.bListGetTime
                {
                                                        //********* Adding or Reading **************
                    _OS_temp = _fsr;        // Store pointer to current timer->Next
                    _fsr = _indf + 2;       // Fsr -> fsr->Next->Timer

                    //------------------------------------------------------------------------------
                    // Copy time value into temp variable

                    *((unsigned char *)&temp + 0) = _indf;
                    #if   OS_QTIMER_SIZE > 1
                        _fsr++; *((unsigned char *)&temp + 1) = _indf;
                    #if   OS_QTIMER_SIZE > 2
                        _fsr++; *((unsigned char *)&temp + 2) = _indf;
                        _fsr++; *((unsigned char *)&temp + 3) = _indf;
                    #endif
                    #endif

                    _fsr = _OS_temp;    // fsr points to ftcb->Next

                    if (_OS_QTimers.Flags.bListGetTime)
                    {
                        if ((unsigned char)(_fsr-1) == (_FSR_TYPE)ftimer) goto EXIT_RI;
                        remain -= temp;
                    } 
                    else
                    {
                        if (temp && temp <= time)
                        {
                            _fsr--;
                            break;
                        } 
                    }
                }
    
                //------------------------------------------------------------------------------
                // Now FSR points to ftcb->next
                //------------------------------------------------------------------------------
                if (_temp_i & 0x80) _gie = 1;
                time -= temp;
                _fsr = _indf;
                di();
    
            } while (_OS_IndfFTimerNextEnable);
        }
    
        //**************************************
        //               Adding
        //**************************************
        if (_OS_QTimers.Flags.bListFunction == OS_QTIMER_LIST_FUNCTION_ADD)
        {
            *((unsigned char *)&remain+0) = _fsr;   // Store &ftcb
            *((unsigned char *)&flags)    = _indf;  // store "flags"
            _fsr++;
            *((unsigned char *)&temp + 0) = _indf;  // Store ftcb->Next
            _fsr = ((_FSR_TYPE)ftimer);             // fsr = ftimer
            _OS_IndfFTimerActive        = 1;
            _OS_IndfFTimerTimeout       = 0;
            _fsr++;
            _indf = *((unsigned char *)&temp + 0);  // ftimer->Next = stored "Next"
    
            _fsr++;                                 // ftimer->Timer = time
            //------------------------------------------------------------------------------
            _indf = *((unsigned char *)&time + 0);
        #if     OS_QTIMER_SIZE > 1
            _fsr++; _indf = *((unsigned char *)&time + 1);
        #if     OS_QTIMER_SIZE > 2
            _fsr++; _indf = *((unsigned char *)&time + 2);
            _fsr++; _indf = *((unsigned char *)&time + 3);
        #endif
        #endif
            //------------------------------------------------------------------------------
    
            _fsr = *((unsigned char *)&remain + 0);     // Restore: _fsr = ftcb
            if (flags.bTimeout) _fsr = (_FSR_TYPE)&_OS_QTimers;
    
            _OS_IndfFTimerNextEnable = 1;               // ftcb->Next = ftimer
            _fsr++;
            _indf = (unsigned char) ftimer;
    
            _fsr = (_FSR_TYPE)ftimer;
    
            if (flags.bNextEnable)               // Decrement next timer: ftimer->Next->Timer -= time
            {
                _OS_IndfFTimerNextEnable = 1;
                _fsr = *((unsigned char *)&temp + 0) + 2;    // _fsr = &ftimer->Next->Timer
        
                //------------------------------------------------------------------------------
                _indf -= *((unsigned char *)&time + 0);
        
                #if   OS_QTIMER_SIZE == 2
        
                    _fsr++;
                    if (!CARRY) _indf--;
                    _indf -= *((unsigned char*)&time + 1);

                #endif
                            
                #if   OS_QTIMER_SIZE == 4
        
                    if (!CARRY) time += 0x100;
                    _fsr++;
                    _indf -= *((unsigned char *)&time + 1);
                    if (!CARRY) time += 0x10000;
                    _fsr++;
                    _indf -= *((unsigned char *)&time + 2);
                    if (!CARRY) time += 0x1000000;
                    _fsr++;
                    _indf -= *((unsigned char *)&time + 3);
        
                #endif
                //------------------------------------------------------------------------------
            }
    
    
        }
    
    EXIT_RI:
        if (_temp_i & 0x80) _gie = 1;
    EXIT:
        _OS_QTimers.Flags.bListFunction = OS_QTIMER_LIST_FUNCTION_ADD;
        _OS_QTimers.Flags.bListGetTime  = 0;
        return remain;
    }
    


//******************************************************************************
//#elif defined(__OSAPICC18__)
//******************************************************************************



//******************************************************************************
#else
//******************************************************************************

_OST_QTIMER _OS_Qtimer_List (_OST_QTIMER time, OST_QTIMER *ftimer)
{
    OST_QTIMER          *ftcb;
    OST_QTIMER_FLAGS    flags;
    _OST_QTIMER         temp, remain;
    __osa_word       itemp;
    #ifdef __OSAMIKROC18__
    OST_QTIMER          *ftcb_temp; // Due to compiler bug (id=4711) can't copy pointed value
                                    // (*Note:4711*) (version mikroC PRO 3.20)
    #endif

RESTART:
    remain = 0;                                 // Count remaining time
    flags = ftimer->Flags;

    if (_OS_QTimers.Flags.bListGetTime) 
    {    
        //--------------------------------------------
        // Clear ADD function if GET function present
        //--------------------------------------------
        _OS_QTimers.Flags.bListFunction = !OS_QTIMER_LIST_FUNCTION_ADD;
    }
    else 
    {
        //---------------------------------------
        // Skip deleting if timer is not in list
        //---------------------------------------
        if (_OS_QTimers.Flags.bListFunction == OS_QTIMER_LIST_FUNCTION_DELETE && !flags.bActive) goto EXIT;
    }

    _OS_QTimers.Flags.bAddAfterDelete = 0;

    if (_OS_QTimers.Flags.bListFunction == OS_QTIMER_LIST_FUNCTION_ADD)
    {
        if (flags.bActive)              // If timer is already in list, it should be deleted first
        {
            OS_QTIMER_ATOMIC_WRITE_A(_OS_QTimers.Flags.bAddAfterDelete = 1;);  // Set flag to execute ADD after deleting
            _OS_QTimers.Flags.bListFunction = OS_QTIMER_LIST_FUNCTION_DELETE;
        }
        else
        {
            //---------------------------------------
            // Prepare timer for adding
            //---------------------------------------
            OS_QTIMER_ATOMIC_WRITE_A(ftimer->Flags.bNextEnable = 0;);
            time ^= (_OST_QTIMER)-1; // time = -time
            time++;
        }
    }

    #if defined(__OSAMCC30__)
    ftcb = (OST_QTIMER*)(void*)&_OS_QTimers;    // 
    #else
    ftcb = (OST_QTIMER*)&_OS_QTimers;           // Start search from header
    #endif

    itemp = OS_DI();

    if (_OS_QTimers.Flags.bNextEnable)
    {    
        do
        {
            #ifdef __OSAMIKROC18__          // (*Note:4711*)
            ftcb_temp = ftcb->Next;         // (*Note:4711*)
            #endif                          // (*Note:4711*)

            if (_OS_QTimers.Flags.bListFunction == OS_QTIMER_LIST_FUNCTION_DELETE && !_OS_QTimers.Flags.bListGetTime)
            {
                if (ftcb->Next == ftimer)               //********* Deleting ***********
                {
                    ftcb->Next = ftimer->Next;
                    OS_QTIMER_ATOMIC_WRITE_A(
                        ftcb->Flags.bNextEnable = flags.bNextEnable;
                        ftimer->Flags.bActive   = 0;
                    );

                    if (flags.bNextEnable) ftcb->Next->Timer += ftimer->Timer;

                    if (_OS_QTimers.Flags.bAddAfterDelete)
                    {
                        //-----------------------------------------------------------
                        // It was deleting before adding. 
                        // Set ADD flag and restart searching
                        //-----------------------------------------------------------

                        _OS_QTimers.Flags.bListFunction = OS_QTIMER_LIST_FUNCTION_ADD;
                        goto RESTART;
                    }
                    goto EXIT_RI;
                }
            }
            else                                        //********* Adding or Reading ***********
            {
                #ifdef __OSAMIKROC18__      // (*Note:4711*)
                temp = ftcb_temp->Timer;    // (*Note:4711*)
                #else                       // (*Note:4711*)
                temp = ftcb->Next->Timer;   // Copy time value into temp variable
                #endif                      // (*Note:4711*)

                if (_OS_QTimers.Flags.bListGetTime)
                {
                    if (ftcb == ftimer) goto EXIT_RI;
                    remain -= temp;
                } 
                else
                {
                    if (temp && temp <= time)
                    {
                        OS_QTIMER_ATOMIC_WRITE_A(ftimer->Flags.bNextEnable = 1);
                        break;
                    } 
                }
                time -= temp;
            }

            OS_RI(itemp);
            ftcb = ftcb->Next;
            OS_DI();

        } while (ftcb->Flags.bNextEnable);
    }

    //**************************************
    //               Adding
    //**************************************
    if (_OS_QTimers.Flags.bListFunction == OS_QTIMER_LIST_FUNCTION_ADD)
    {
        OS_QTIMER_ATOMIC_WRITE_A(
            ftimer->Flags.bTimeout = 0;
            ftimer->Flags.bActive  = 1;
        );
        ftimer->Next = ftcb->Next;
        ftimer->Timer = time;
    
        if (ftcb->Flags.bTimeout)
        {
            #if defined(__OSAMCC30__)
            ftcb = (OST_QTIMER*)(void*)&_OS_QTimers;    // 
            #else
            ftcb = (OST_QTIMER*)&_OS_QTimers;           // Start search from header
            #endif
        }
        ftcb->Next = ftimer;
        OS_QTIMER_ATOMIC_WRITE_A(ftcb->Flags.bNextEnable = 1);
    
        if (ftimer->Flags.bNextEnable)
        {
            ftimer = ftimer->Next;
            ftimer->Timer -= time;
        }
    }

EXIT_RI:
    OS_RI(itemp);
EXIT:
    _OS_QTimers.Flags.bListFunction = OS_QTIMER_LIST_FUNCTION_ADD;
    _OS_QTimers.Flags.bListGetTime  = 0;
    return remain;
}

//******************************************************************************
#endif      // type of compiler
//******************************************************************************

#endif  // OS_ENABLE_QTIMERS



